# !/usr/bin/python
# coding=utf-8
from uitk import Signals


class ExampleSlots:
    def __init__(self, *args, **kwargs):
        self.sb = self.switchboard()

    def header_init(self, widget):
        widget.configureButtons(
            menu_button=True, minimize_button=True, hide_button=True
        )
        widget.menu.setTitle("DRAG ME!")
        widget.menu.add(self.sb.Label, setObjectName="label", setText="Clickable Label")

    def label(self):
        widget = self.sb.example.header.menu.label
        self.sb.message_box(f"{widget.text()}: Clicked")

    def button_a(self, widget):
        self.sb.message_box(f"{widget.text()}: Clicked")

    def button_b_init(self, widget):
        widget.menu.setTitle("OPTION MENU")
        widget.menu.add(
            "QRadioButton", setObjectName="radio_a", setText="Option A", setChecked=True
        )
        widget.menu.add("QRadioButton", setObjectName="radio_b", setText="Option B")
        widget.menu.add("QRadioButton", setObjectName="radio_c", setText="Option C")

    @Signals("released")
    def button_b(self, widget):
        option = (
            "A"
            if widget.menu.radio_a.isChecked()
            else "B"
            if widget.menu.radio_b.isChecked()
            else "C"
        )
        self.sb.message_box(f"{widget.text()}: Option {option}")

    def checkbox(self, state):
        self.sb.message_box(
            f'CheckBox State: <hl style="color:yellow;">{bool(state)}</hl>'
        )

    def spinbox(self, value, widget):
        self.sb.message_box(f'SpinBox Value: <b style="font-weight: bold;">{value}</b>')

    @staticmethod
    def textedit_init(widget):
        widget.setText(
            "• Collapse this text edit by clicking on the three dots above it.\n\n• Widget states are restored the next time this UI is opened."
        )


# --------------------------------------------------------------------------------------------

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtWidgets
from uitk.widgets.mixins.attributes import AttributesMixin


class MessageBox(QtWidgets.QMessageBox, AttributesMixin):
    """Displays a message box with HTML formatting for a set time before closing.

    Parameters:
        location (str)(point) = move the messagebox to the specified location. Can be given as a qpoint or string value. default is: 'topMiddle'
        timeout (int): time in seconds before the messagebox auto closes.
    """

    def __init__(self, parent=None, location="topMiddle", timeout=2, **kwargs):
        QtWidgets.QMessageBox.__init__(self, parent)

        self.setWindowModality(QtCore.Qt.NonModal)
        self.setStandardButtons(QtWidgets.QMessageBox.NoButton)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.setWindowFlags(
            QtCore.Qt.WindowType.WindowDoesNotAcceptFocus
            | QtCore.Qt.WindowStaysOnTopHint
            | QtCore.Qt.Tool
            | QtCore.Qt.FramelessWindowHint
        )

        self.menu_timer = QtCore.QTimer()
        self.menu_timer.setSingleShot(True)
        self.menu_timer.timeout.connect(self.hide)

        self.setTextFormat(QtCore.Qt.RichText)

        self.location = location

        self.set_attributes(**kwargs)

    def move_(self, location) -> None:
        # Get the screen's geometry
        screen = QtWidgets.QApplication.screens()[0]
        rect = screen.geometry()

        offset_x = self.sizeHint().width() / 2
        offset_y = self.sizeHint().height() / 2

        if location == "topMiddle":
            point = QtCore.QPoint(rect.width() / 2 - offset_x, rect.top() + 150)
        elif location == "bottomRight":
            point = QtCore.QPoint(rect.width() - offset_x, rect.height() - offset_y)
        elif location == "topLeft":
            point = QtCore.QPoint(rect.left() + offset_x, rect.top() + offset_y)
        elif location == "bottomLeft":
            point = QtCore.QPoint(rect.left() + offset_x, rect.height() - offset_y)
        else:  # default to the middle of the screen if location is not recognized
            point = QtCore.QPoint(
                rect.width() / 2 - offset_x, rect.height() / 2 - offset_y
            )

        self.move(point)

    def _setPrefixStyle(self, string) -> str:
        """Set style for specific keywords in the given string.

        Returns:
                (str)
        """
        style = {
            "Error:": '<hl style="color:red;">Error:</hl>',
            "Warning:": '<hl style="color:yellow;">Warning:</hl>',
            "Note:": '<hl style="color:blue;">Note:</hl>',
            "Result:": '<hl style="color:green;">Result:</hl>',
        }

        for k, v in style.items():
            string = string.replace(k, v)

        return string

    def _setHTML(self, string) -> str:
        """<p style="font-size:160%;">text</p>
        <p style="text-align:center;">Centered paragraph.</p>
        <p style="font-family:courier;">This is a paragraph.</p>

        Returns:
                (str)
        """
        style = {
            "<p>": '<p style="color:white;">',  # paragraph <p>' </p>'
            "<hl>": '<hl style="color:yellow; font-weight: bold;">',  # heading <h1>' </h1>'
            "<body>": '<body style="color;">',  # body <body> </body>
            "<b>": '<b style="font-weight: bold;">',  # bold <b> </b>
            "<strong>": '<strong style="font-weight: bold;">',  # <strong> </strong>
            "<mark>": '<mark style="background-color: grey">',  # highlight <mark> </mark>
        }

        for k, v in style.items():
            string = string.replace(k, v)

        return string

    def _setFontColor(self, string, color) -> str:
        """
        Returns:
                (str)
        """
        return "<font color=" + color + ">" + string + "</font>"

    def _setBackgroundColor(self, string, color):
        """
        Returns:
                (str)
        """
        return '<mark style="background-color:' + color + '">' + string + "</mark>"

    def _setFontSize(self, string, size) -> str:
        """
        Returns:
                (str)
        """
        return "<font size=" + str(size) + ">" + string + "</font>"

    def setText(
        self, string, fontColor="white", backgroundColor="rgb(50,50,50)", fontSize=5
    ) -> None:
        """Set the text to be displayed.

        Parameters:
                fontColor (str): text color.
                backgroundColor (str): text background color.
                fontSize (int): text size.
        """
        s = self._setPrefixStyle(string)
        s = self._setHTML(s)
        s = self._setFontColor(s, fontColor)
        s = self._setBackgroundColor(s, backgroundColor)
        s = self._setFontSize(s, fontSize)

        super().setText(s)

    def showEvent(self, event) -> None:
        """ """
        self.menu_timer.start(1000)  # 5000 milliseconds = 5 seconds
        self.move_(self.location)

        super().showEvent(event)

    def hideEvent(self, event) -> None:
        """ """
        self.menu_timer.stop()

        super().hideEvent(event)


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    # Return the existing QApplication object, or create a new one if none exists.
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    w = MessageBox()
    w.setText("Warning: Backface Culling is now <hl>Off</hl>")
    w.show()

    sys.exit(app.exec_())

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace, 
        then right-click them and select 'Promote to...'. 

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote", 
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""

# deprecated: -----------------------------------


# class ShowMessageBox(MessageBox):
#     """Spawns a message box with the given text.
#     Supports HTML formatting.
#     Prints a formatted version of the given string to console, stripped of html tags, to the console.

#     Parameters:
#         message_type (str/optional): The message context type. ex. 'Error', 'Warning', 'Info', 'Result'
#         location (str/QPoint/optional) = move the messagebox to the specified location. default is: 'topMiddle'
#         timeout (int/optional): time in seconds before the messagebox auto closes. default is: 3
#     """

#     def __init__(
#         self,
#         string,
#         message_type="",
#         location="topMiddle",
#         timeout=3,
#         **kwargs,
#     ):
#         super().__init__(
#             QtWidgets.QApplication.instance(), **kwargs
#         )  # Set the QApplication instance as the parent

#         if message_type:
#             string = f"{message_type.capitalize()}: {string}"

#         self.location = location
#         self.timeout = timeout

#         self.setText(string)
#         self.setVisible(True)

#         # strip everything between '<' and '>' (html tags)
#         print(f"# {re.sub('<.*?>', '', string)}")

#     def hideEvent(self, event) -> None:
#         """ """
#         self.menu_timer.stop()

#         super().hideEvent(event)


# def insertText(self, dict_):
#   '''
#   Parameters:
#       dict_ = {dict} - contents to add.  for each key if there is a value, the key and value pair will be added.
#   '''
#   highlight = QtGui.QColor(255, 255, 0)
#   baseColor = QtGui.QColor(185, 185, 185)

#   #populate the textedit with any values
#   for key, value in dict_.items():
#       if value:
#           self.setTextColor(baseColor)
#           self.append(key) #textEdit.append(key+str(value))
#           self.setTextColor(highlight)
#           self.insertPlainText(str(value))



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtGui, QtWidgets
from uitk.widgets.mixins.attributes import AttributesMixin
from uitk.widgets.mixins.text import RichText, TextOverlay


class Header(QtWidgets.QLabel, AttributesMixin, RichText, TextOverlay):
    """Header is a QLabel that can be dragged around the screen and can be pinned/unpinned. It provides a customizable
    header bar with buttons for common window actions such as minimizing, hiding, and pinning.

    Signals:
        toggled(bool): Emitted when the pin state is toggled.

    Attributes:
        button_definitions (dict): Defines the properties of the buttons available in the header.
        state (str): Represents the current state of the header ("unpinned", "pinned").
    """

    toggled = QtCore.Signal(bool)

    # Define button properties
    button_definitions = {
        "menu_button": ("≡", "show_menu"),
        "minimize_button": ("–", "minimize_window"),
        "hide_button": ("×", "hide_window"),
        "pin_button": ("\u25CB", "toggle_pin"),
    }

    def __init__(
        self,
        parent=None,
        **kwargs,
    ):
        super().__init__(parent)
        """Initialize the Header with buttons and layout.

        Parameters:
            parent (QWidget, optional): The parent widget. Defaults to None.
            **kwargs: Additional keyword arguments to configure buttons and other attributes.
                      Accepts keys corresponding to the button_definitions, and any additional attributes.
        """
        self.pinned = False  # unpinned, pinned
        self.__mousePressPos = None

        self.container_layout = QtWidgets.QHBoxLayout(self)
        self.container_layout.setContentsMargins(0, 0, 0, 0)
        self.container_layout.setSpacing(1)
        self.container_layout.addStretch(1)

        self.setLayout(self.container_layout)
        self.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        self.setStyleSheet(self.getStyleSheet())

        # Extract button-related arguments
        button_args = {
            key: kwargs.pop(key)
            for key in list(kwargs.keys())
            if key in self.button_definitions
        }

        self.setFixedHeight(20)
        self.configureButtons(**button_args)
        self.set_attributes(**kwargs)

    def getButtonStyleSheet(self):
        """Return the stylesheet for buttons."""
        return """
            QPushButton { background-color: transparent; border: none;}
            QPushButton:hover { background-color: rgba(127,127,127,200); border: none;}
            QPushButton#hide_button:hover { background-color: rgba(255,0,0,200); border: none;}
        """

    def getStyleSheet(self):
        """Return the stylesheet for the header label."""
        return """
            QLabel {
                background-color: rgba(127,127,127,2);
                border: none;
                font-weight: bold;
            }
            QLabel::hover {
                background-color: rgba(127,127,127,2);
                border: none;
            }
        """

    def createButton(self, text, callback, button_type=None):
        """Create a button with the given text and callback."""
        button = QtWidgets.QPushButton(text, self)
        if button_type:
            button.setObjectName(button_type)
        button.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
        button.setStyleSheet(self.getButtonStyleSheet())
        button.clicked.connect(callback)
        return button

    def configureButtons(self, **kwargs):
        """Configure buttons based on the given parameters."""
        # Clear existing buttons from the layout
        for i in reversed(range(self.container_layout.count())):
            widget = self.container_layout.itemAt(i).widget()
            if isinstance(widget, QtWidgets.QPushButton):
                self.container_layout.removeWidget(widget)

        # Create and add buttons to the layout based on parameters
        self.buttons = {}
        for param, visible in kwargs.items():
            if param in self.button_definitions and visible:
                text, method_name = self.button_definitions[param]
                callback = getattr(self, method_name)
                button = self.createButton(text, callback, button_type=param)
                self.container_layout.addWidget(button)
                button.show()
                self.buttons[param] = button

        # Ensure layout is updated
        self.container_layout.invalidate()
        self.triggerResizeEvent()

    def triggerResizeEvent(self):
        current_size = self.size()
        resize_event = QtGui.QResizeEvent(current_size, current_size)
        self.resizeEvent(resize_event)

    def resizeEvent(self, event):
        self.resizeButtons()
        self.updateFontSize()
        super().resizeEvent(event)

    def resizeButtons(self):
        button_size = self.height()
        margin = button_size * 0.05
        for button_name, button in self.buttons.items():
            button.setFixedSize(button_size - margin, button_size - margin)

    def updateFontSize(self):
        # Calculate font size for the label and buttons relative to widget's height
        label_font_size = self.height() * 0.4
        button_font_size = self.height() * 0.6  # 60% of the widget's height

        # Apply font size to the label
        label_font = self.font()
        label_font.setPointSizeF(label_font_size)
        self.setFont(label_font)

        # Iterate through the widgets in the layout and update the font size for the buttons
        for i in range(self.container_layout.count()):
            widget = self.container_layout.itemAt(i).widget()
            if isinstance(widget, QtWidgets.QPushButton):
                button_font = widget.font()
                button_font.setPointSizeF(button_font_size)
                widget.setFont(button_font)

    def setTitle(self, title):
        """Set the title of the header.

        Parameters:
            title (str): The new title.
        """
        self.setText(title)

    def title(self):
        """Get the title of the header.

        Returns:
            str: The current title.
        """
        return self.text()

    def minimize_window(self):
        """Minimize the parent window."""
        self.window().showMinimized()

    def hide_window(self):
        """Hide the parent window."""
        if self.pinned:
            self.toggle_pin()
        self.window().hide()

    @property
    def menu(self):
        try:
            return self._menu
        except AttributeError:
            from uitk.widgets.menu import Menu

            self._menu = Menu(self, fixed_item_height=20)
            return self._menu

    def show_menu(self):
        """Show the menu."""
        self.menu.setVisible(True)

    def toggle_pin(self):
        """Toggle pinning of the window."""
        state = not self.pinned

        self.pinned = state
        self.window().prevent_hide = state
        pin_button_text = "\u25CF" if state else "\u25CB"

        pin_button = self.buttons.get("pin_button")
        if pin_button:
            pin_button.setText(pin_button_text)
            if not state:
                self.window().hide()

        self.toggled.emit(state)

    def mousePressEvent(self, event):
        """Handle the mouse press event. If the left button is pressed, store the global position of the mouse cursor.

        Parameters:
            event (QMouseEvent): The mouse event.
        """
        if event.button() == QtCore.Qt.LeftButton:
            self.__mousePressPos = event.globalPos()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """Handle the mouse move event."""
        if self.__mousePressPos is not None:
            moveAmount = event.globalPos() - self.__mousePressPos
            if moveAmount.manhattanLength() > 5:
                self.window().move(self.window().pos() + moveAmount)
                self.__mousePressPos = event.globalPos()
                if not self.pinned:  # Only change state if not already pinned
                    self.toggle_pin()
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        # No need to toggle pin state here, as pinning is controlled by the button
        self.__mousePressPos = None
        super().mouseReleaseEvent(event)

    def showEvent(self, event):
        """Show the menu button if it exists and contains items."""
        menu_button = self.buttons.get("menu_button")
        if menu_button:
            menu_button.setVisible(self.menu.contains_items)

        super().showEvent(event)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    # return the existing QApplication object, or create a new one if none exists.
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    w = QtWidgets.QWidget()
    layout = QtWidgets.QVBoxLayout(w)
    header = Header(
        menu_button=True,
        minimize_button=True,
        pin_button=True,
        hide_button=True,
        setTitle="DRAG ME!",
    )
    header.toggled.connect(lambda state: print(f"Header pinned: {state}!"))
    header.menu.add(["Menu Item A", "Menu Item B"])

    layout.addWidget(header)
    w.show()

    sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace,
        then right-click them and select 'Promote to...'.

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote",
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""



# !/usr/bin/python
# coding=utf-8
from functools import wraps


class Signals:
    """Class-based decorator to annotate slot methods with the signals to which they should connect.

    This class takes one or more signal names as strings during initialization and assigns them as attributes
    to the decorated function. The signals can be later retrieved for connecting the slot method to the respective signals.

    Attributes:
        signals (tuple of str): Tuple containing one or more signal names as strings.

    Raises:
        ValueError: If no signals are provided during initialization.
        TypeError: If any of the provided signals is not a string.

    Example:
        @Signals('clicked', 'pressed')
        def on_button_interaction():
            print("Button interacted")
    """

    def __init__(self, *signals):
        if len(signals) == 0:
            raise ValueError("At least one signal must be specified")

        for signal in signals:
            if not isinstance(signal, str):
                raise TypeError(f"Signal must be a string, not {type(signal)}")

        self.signals = signals

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper.signals = self.signals
        return wrapper



# !/usr/bin/python
# coding=utf-8
import re
import logging
from PySide2 import QtCore, QtGui, QtWidgets


class EventFactoryFilter(QtCore.QObject):
    """Event filter for dynamic UI objects.
    Forwards events to event handlers dynamically based on the event type.

    Parameters:
        parent (QWidget, optional): The parent widget for the event filter. Defaults to None.
    """

    def __init__(self, parent=None, forward_events_to=None, event_name_prefix=""):
        super().__init__(parent)

        self.forward_events_to = forward_events_to or self
        self.event_name_prefix = event_name_prefix

    @staticmethod
    def format_event_name(event_type, prefix=""):
        """Get a formatted event method name string from a given event type using a regular expression.

        Parameters:
            event_type (QEvent.Type): The event type whose method name needs to be generated.
            prefix (str, optional): A prefix for the event method names. Defaults to an empty string.

        Returns:
            str: The formatted event method name.

        Examples:
            format_event_name(QtCore.QEvent.Type.Enter) returns 'enterEvent'
            format_event_name(QtCore.QEvent.Type.MouseButtonPress) returns 'mousePressEvent'
            format_event_name(QtCore.QEvent.Type.Enter, prefix='ef_') returns 'ef_enterEvent'
            format_event_name(QtCore.QEvent.Type.MouseButtonPress, prefix='ef_') returns 'ef_mousePressEvent'
        """
        event_name = re.sub(
            r"^.*\.([A-Z])([^B]*)(?:Button)?(.*)$",
            lambda m: prefix + m.group(1).lower() + m.group(2) + m.group(3) + "Event",
            str(event_type),
        )
        return event_name

    def eventFilter(self, widget, event):
        """Forward widget events to event handlers.
        For any event type, the eventFilter will try to connect to a corresponding
        method derived from the event type string.

        Parameters:
            widget (QWidget): The widget that the event filter is applied to.
            event (QEvent): The event that needs to be processed.

        Returns:
            bool: True if the event was handled, False otherwise.
        """
        try:
            event_handler = getattr(
                self.forward_events_to,
                self.format_event_name(event.type(), self.event_name_prefix),
            )
            event_handled = event_handler(widget, event)
            if event_handled:
                return True
        except AttributeError:
            pass

        return False


class MouseTracking(QtCore.QObject):
    """MouseTracking is a QObject subclass that provides mouse enter and leave events for QWidget child widgets.
    It uses event filtering to track the mouse movement and send enter and leave events to the child widgets.

    Attributes:
        _prev_mouse_over (list): List of widgets that were previously under the mouse cursor.
        _mouse_over (list): List of widgets that are currently under the mouse cursor.
        _filtered_widgets (set): Set of widgets that have been processed for special handling (widgets with a viewport).
        logger (Logger): Instance of a logger that logs mouse tracking events.

    Methods:
        eventFilter(self, widget, event): Filters events to track mouse move events and button press/release events.
        should_capture_mouse(self, widget): Checks if a widget should capture the mouse.
        track(self): Updates tracking data and sends enter, leave, and release events to widgets.

    Parameters:
        parent (QWidget): Parent widget for the MouseTracking object.
        log_level (int, optional): Logging level. Defaults to logging.WARNING.

    Raises:
        TypeError: If parent is not a QWidget derived type.
    """

    def __init__(self, parent, log_level=logging.WARNING):
        super().__init__(parent)

        if not isinstance(parent, QtWidgets.QWidget):
            raise TypeError("Parent must be a QWidget derived type")

        self._initialize_logger(log_level)
        self._prev_mouse_over = []
        self._mouse_over = []
        self._filtered_widgets = set()

        parent.installEventFilter(self)

    def _initialize_logger(self, log_level):
        """Initializes logger."""
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(log_level)
        handler = logging.StreamHandler()
        handler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        self.logger.addHandler(handler)

    def should_capture_mouse(self, widget):
        """Checks if a widget should capture the mouse."""
        widget_conditions = [
            (QtWidgets.QPushButton, lambda widget: not widget.isDown()),
            (QtWidgets.QComboBox, lambda widget: not widget.view().isVisible()),
            (QtWidgets.QSlider, lambda widget: not widget.isSliderDown()),
            (QtWidgets.QScrollBar, lambda widget: not widget.isSliderDown()),
        ]
        for widget_type, condition in widget_conditions:
            if isinstance(widget, widget_type) and condition(widget):
                self.logger.debug(
                    f"Not capturing mouse for {widget_type.__name__} under specified condition"
                )
                return False
        return True

    def _update_widgets_under_cursor(self):
        """Updates the list of widgets currently under the cursor."""
        self._get_child_widgets()
        top_widget = QtWidgets.QApplication.widgetAt(QtGui.QCursor.pos())
        self._mouse_over = (
            [top_widget] if top_widget and top_widget in self._widgets else []
        )
        self.logger.debug(
            f"Widgets under cursor: {[f'{w}, {type(w)}' for w in self._mouse_over]}"
        )

    def _get_child_widgets(self):
        """Updates the list of child widgets of the parent."""
        parent = self.parent()
        self._widgets = (
            parent.currentWidget().findChildren(QtWidgets.QWidget)
            if isinstance(parent, QtWidgets.QStackedWidget)
            else parent.findChildren(QtWidgets.QWidget)
        )

    def track(self):
        """Updates tracking data and sends enter and leave events to widgets."""
        self.logger.info(f"Previous widgets under cursor: {self._prev_mouse_over}")
        self.logger.info(f"Current widgets under cursor: {self._mouse_over}")
        self._release_mouse_for_widgets(self._mouse_over)
        self._update_widgets_under_cursor()

        for widget in self._prev_mouse_over:
            if widget not in self._mouse_over:
                self._send_leave_event(widget)
        for widget in self._mouse_over:
            if widget not in self._prev_mouse_over:
                self._send_enter_event(widget)

        self._handle_mouse_grab()
        self._prev_mouse_over = self._mouse_over.copy()
        self._filter_viewport_widgets()

    def _release_mouse_for_widgets(self, widgets):
        """Releases mouse for given widgets."""
        for widget in widgets:
            widget.releaseMouse()

    def _send_leave_event(self, widget):
        """Sends a leave event to a widget."""
        self.logger.info(
            f"Sending Leave event to: {widget}, Name: {widget.objectName()}, Parent: {widget.parent().objectName()}"
        )
        QtGui.QGuiApplication.sendEvent(widget, QtCore.QEvent(QtCore.QEvent.Leave))

    def _send_enter_event(self, widget):
        """Sends an enter event to a widget."""
        self.logger.info(
            f"Sending Enter event to: {widget}, Name: {widget.objectName()}, Parent: {widget.parent().objectName()}"
        )
        QtGui.QGuiApplication.sendEvent(widget, QtCore.QEvent(QtCore.QEvent.Enter))

    def _send_release_event(self, widget, button):
        """Sends a release event to a widget."""
        self.logger.info(
            f"Sending Release event to: {widget}, Name: {widget.objectName()}, Parent: {widget.parent().objectName()}"
        )
        release_event = QtGui.QMouseEvent(
            QtCore.QEvent.MouseButtonRelease,
            QtGui.QCursor.pos(),
            button,
            button,
            QtCore.Qt.NoModifier,
        )
        QtGui.QGuiApplication.postEvent(widget, release_event)

    def _handle_mouse_grab(self):
        """Handles mouse grabbing depending on the widget currently under the cursor."""
        top_widget = QtWidgets.QApplication.widgetAt(QtGui.QCursor.pos())
        if top_widget:
            self.logger.debug(f"Top widget under cursor: {top_widget}")
            widget_to_grab = (
                top_widget
                if self.should_capture_mouse(top_widget)
                else QtWidgets.QApplication.activeWindow()
            )
            self.logger.info(f"Grabbing mouse for widget: {widget_to_grab}")
            widget_to_grab.grabMouse()
        else:
            self.logger.debug(
                "No widget under cursor. Grabbing mouse for active window."
            )
            QtWidgets.QApplication.activeWindow().grabMouse()

    def _filter_viewport_widgets(self):
        """Adds special handling for widgets with a viewport."""
        for widget in self._widgets:
            if hasattr(widget, "viewport") and widget not in self._filtered_widgets:
                self._filtered_widgets.add(widget)
                self._handle_viewport_widget(widget)

    def _handle_viewport_widget(self, widget):
        """Ignores mouse move events for widgets with a viewport."""
        original_mouse_move_event = widget.mouseMoveEvent
        widget.mouseMoveEvent = lambda event: (
            original_mouse_move_event(event),
            event.ignore(),
        )

    def eventFilter(self, widget, event):
        """Calls `track` on each mouse move event and also tracks button press/release for QAbstractButton."""
        if event.type() == QtCore.QEvent.MouseMove:
            self.logger.info(
                f"MouseMove event filter triggered by: {widget} with event: {event.type()}"
            )
            self.track()

        elif event.type() == QtCore.QEvent.MouseButtonRelease:
            top_widget = QtWidgets.QApplication.widgetAt(QtGui.QCursor.pos())
            if (
                isinstance(top_widget, QtWidgets.QAbstractButton)
                and not top_widget.isDown()
            ):
                self.logger.info(
                    f"Mouse button release event detected on: {top_widget}"
                )
                self._send_release_event(top_widget, event.button())

        return super().eventFilter(widget, event)


# --------------------------------------------------------------------------------------------

logging.info(__name__)  # module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



import sys
from PySide2.QtCore import QObject, Qt, QEvent, QTimer, Signal, QRect
from PySide2.QtGui import QPainter, QColor, QBrush
from PySide2.QtWidgets import (
    QApplication,
    QMainWindow,
    QDockWidget,
    QVBoxLayout,
    QWidget,
    QLabel,
)


class DockingOverlay(QWidget):
    def __init__(self, docking):
        super().__init__()
        self.docking = docking
        self.setAttribute(Qt.WA_TransparentForMouseEvents, True)
        self.setAttribute(Qt.WA_TranslucentBackground, True)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)

        self.docking.dock_position_changed.connect(self.update)
        self.raise_()

    def update(self):
        dock_position = self.docking.dock_position
        dock_window = (
            self.docking.dock_window
        )  # Get the reference to the window being dragged over
        if not dock_position or not dock_window:
            self.hide()
            return

        self.setGeometry(
            dock_window.geometry()
        )  # Use the geometry of the window being dragged over
        self.show()

    def paintEvent(self, event):
        position = self.docking.dock_position
        if not position:
            return

        painter = QPainter(self)
        color = QColor(0, 255, 0, 100)  # Green color with transparency
        painter.setBrush(QBrush(color))
        painter.setPen(Qt.NoPen)

        rect_width = 10
        rect_height = self.height()

        if position == "right":
            rect_x = 0
            rect_y = 0
        elif position == "left":
            rect_x = self.width() - rect_width
            rect_y = 0
        elif position == "top":
            rect_x = 0
            rect_y = self.height() - 10
            rect_width = self.width()
            rect_height = 10
        elif position == "bottom":
            rect_x = 0
            rect_y = 0
            rect_width = self.width()
            rect_height = 10
        else:
            return

        painter.drawRect(rect_x, rect_y, rect_width, rect_height)


class DockingWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("DockingMixin Window")
        self.setGeometry(300, 100, 400, 300)
        self.setWindowFlags(Qt.Tool)
        self.docked_widgets = []

    def add_tool_window(self, tool_window, position):
        dock_widget = CustomDockWidget(tool_window.windowTitle(), self)
        dock_widget.setAllowedAreas(Qt.AllDockWidgetAreas)
        dock_widget.setWidget(tool_window)
        self.addDockWidget(position, dock_widget)
        self.docked_widgets.append(dock_widget)
        dock_widget.undocked.connect(self.remove_tool_window)

    def remove_tool_window(self, tool_window):
        if tool_window in self.docked_widgets:
            self.docked_widgets.remove(tool_window)

    def get_docked_widgets(self):
        return self.docked_widgets


class CustomDockWidget(QDockWidget):
    docked = Signal(QDockWidget)
    undocked = Signal(QDockWidget)

    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.setFeatures(
            QDockWidget.DockWidgetMovable | QDockWidget.DockWidgetFloatable
        )
        self.setFloating(False)
        self.topLevelChanged.connect(self.handle_top_level_change)
        self.installEventFilter(self)

    def handle_top_level_change(self, floating):
        if floating:
            self.undocked.emit(self)
        else:
            self.docked.emit(self)

    def eventFilter(self, widget, event):
        if event.type() == QEvent.MouseMove:
            if self.isFloating():
                # Check if any part of the dock widget is inside the docking window
                if self.parent().geometry().intersects(self.geometry()):
                    print("Dragging inside the docking window")
                else:
                    print("Dragging outside the docking window")

                    # Hide the DockingWindow if there are no docked widgets left
                    docked_widgets = self.parent().get_docked_widgets()
                    print(docked_widgets)
                    if not docked_widgets:
                        self.parent().hide()
            else:
                print("Mouse moved while docked")
        return super().eventFilter(widget, event)


class DockingMixin(QObject):
    dock_position_changed = Signal()
    tool_windows = []
    docked_window_groups = []

    def __init__(self, window):
        super().__init__()
        self.window = window
        self._dock_enabled = False
        self._last_dock_position = None
        self.overlay = DockingOverlay(self)
        self.dock_window = None  # Add a reference to the window being docked to

    @property
    def docking_enabled(self):
        return self._dock_enabled

    @docking_enabled.setter
    def docking_enabled(self, value):
        if value == self._dock_enabled:
            return

        self._dock_enabled = value
        if value:
            self.window.installEventFilter(self)
            self.__class__.tool_windows.append(self.window)
        else:
            self.window.removeEventFilter(self)
            self.__class__.tool_windows.remove(self.window)

    @property
    def dock_position(self):
        return self._last_dock_position

    @dock_position.setter
    def dock_position(self, value):
        self._last_dock_position = value
        self.dock_position_changed.emit()

    def dock(self, target_window, position):
        docking_window = DockingWindow()
        self.__class__.docked_window_groups.append(docking_window)
        self.window.setParent(None)
        target_window.setParent(None)
        self.window.setWindowFlags(Qt.Widget)
        target_window.setWindowFlags(Qt.Widget)

        docking_window.add_tool_window(self.window, self.dock_positions[position][0])
        docking_window.add_tool_window(target_window, self.dock_positions[position][1])
        docking_window.show()
        docking_window.activateWindow()

        self.window.removeEventFilter(self)

    @property
    def dock_positions(self):
        return {
            "left": (Qt.RightDockWidgetArea, Qt.LeftDockWidgetArea),
            "right": (Qt.LeftDockWidgetArea, Qt.RightDockWidgetArea),
            "top": (Qt.BottomDockWidgetArea, Qt.TopDockWidgetArea),
            "bottom": (Qt.TopDockWidgetArea, Qt.BottomDockWidgetArea),
        }

    def _get_dock_position(self, window1, window2):
        intersection = window1.geometry().intersected(window2.geometry())

        position_diff = {
            "left": abs(intersection.right() - window1.geometry().left()),
            "right": abs(intersection.left() - window1.geometry().right()),
            "top": abs(intersection.bottom() - window1.geometry().top()),
            "bottom": abs(intersection.top() - window1.geometry().bottom()),
        }
        return min(position_diff, key=position_diff.get)

    def update_docking_position(self):
        dock_distance = 50
        for window in self.__class__.tool_windows:
            if window is not self.window and self.window.geometry().adjusted(
                -dock_distance, -dock_distance, dock_distance, dock_distance
            ).contains(window.geometry().center()):
                dock_position = self._get_dock_position(self.window, window)
                if dock_position:
                    self.dock_position = (
                        dock_position  # Update dock_position with the position only
                    )
                    self.dock_window = window  # Update dock_window reference
                    return
        self.dock_position = None
        self.dock_window = None

    def eventFilter(self, widget, event):
        if not self.docking_enabled:
            return False

        elif event.type() == QEvent.Move:
            self.update_docking_position()

        elif event.type() == QEvent.NonClientAreaMouseButtonRelease:
            if self.dock_position:
                for window in self.__class__.tool_windows:
                    if window is not self.window and self.window.geometry().adjusted(
                        -50, -50, 50, 50
                    ).contains(window.geometry().center()):
                        self.dock(window, self.dock_position)
                        self.dock_position = None
                        break
        return False


if __name__ == "__main__":

    class CustomToolWindow(QMainWindow):
        def __init__(self, title):
            super().__init__()
            self.setWindowFlags(Qt.Tool)
            self.setWindowTitle(title)
            central_widget = QWidget()
            layout = QVBoxLayout(central_widget)
            layout.addWidget(QLabel(title))
            self.setCentralWidget(central_widget)

            self.docking = DockingMixin(self)
            self.docking.docking_enabled = True

    app = QApplication.instance() or QApplication(sys.argv)

    window1 = CustomToolWindow("Window 1")
    window1.setGeometry(100, 100, 200, 200)
    window1.show()

    window2 = CustomToolWindow("Window 2")
    window2.setGeometry(350, 100, 200, 200)
    window2.show()

    sys.exit(app.exec_())



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtGui, QtWidgets
from uitk.widgets.menu import Menu
from uitk.widgets.mixins.attributes import AttributesMixin
from uitk.widgets.mixins.text import RichText, TextOverlay


class CustomListView(QtWidgets.QListView):
    def __init__(self, parent):
        super().__init__(parent)
        self.combobox = parent  # Store a reference to the parent combobox

    def paintEvent(self, event):
        if self.combobox.header_text:  # Check if a header is present
            # Get the index of the first item
            index = self.model().index(0, 0)
            # Get the item rectangle
            rect = self.visualRect(index)
            # Check if the event rect intersects with the first item rectangle
            if event.rect().intersects(rect):
                # Adjust the event rect to exclude the first item rectangle
                event = QtGui.QPaintEvent(event.rect().translated(0, rect.height()))
        super().paintEvent(event)


class AlignedComboBox(QtWidgets.QComboBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.header_alignment = QtCore.Qt.AlignLeft
        self.header_text = None  # New field to store the header text
        self.setView(CustomListView(self))  # Set the custom view

    def setHeader(self, header, alignment="left"):
        self.header_text = header
        self.setHeaderAlignment(alignment)
        self.insertItem(0, header)
        self.setCurrentIndex(0)

    def setHeaderAlignment(self, alignment):
        if isinstance(alignment, str):
            alignment_map = {
                "left": QtCore.Qt.AlignLeft,
                "center": QtCore.Qt.AlignCenter,
                "right": QtCore.Qt.AlignRight,
            }
            self.header_alignment = alignment_map.get(
                alignment.lower(), QtCore.Qt.AlignLeft
            )
        elif isinstance(alignment, QtCore.Qt.AlignmentFlag):
            self.header_alignment = alignment
        else:
            raise ValueError(f"Unsupported alignment type: {type(alignment)}")

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.fillRect(event.rect(), self.palette().base())

        if self.currentIndex() == 0 and self.header_text:
            painter.save()
            rect = event.rect().adjusted(4, 0, 0, 0)  # Adjust margin if needed
            painter.drawText(rect, self.header_alignment, self.currentText())
            painter.restore()

        painter.end()
        super().paintEvent(event)

    def showPopup(self):
        self.model().showingPopup = True  # Set flag before showing popup
        super().showPopup()

    def hidePopup(self):
        self.model().showingPopup = False  # Reset flag after hiding popup
        super().hidePopup()


class ComboBox(AlignedComboBox, AttributesMixin, RichText, TextOverlay):
    """Custom ComboBox widget with additional features and custom signal handling."""

    before_popup_shown = QtCore.Signal()
    on_editing_finished = QtCore.Signal(str)
    on_item_deleted = QtCore.Signal(str)

    def __init__(self, parent=None, editable=False, **kwargs):
        super().__init__(parent)

        self.editable = editable
        self.menu = Menu(self, mode="option", fixed_item_height=20)

        self.set_attributes(**kwargs)

    @property
    def items(self):
        return [
            self.itemData(i) if self.itemData(i) else self.itemText(i)
            for i in range(self.count())
        ]

    def block_signals(fn):
        def wrapper(self, *args, **kwargs):
            self.blockSignals(True)
            rtn = fn(self, *args, **kwargs)
            self.blockSignals(False)
            return rtn

        return wrapper

    def add(
        self,
        x,
        data=None,
        header=None,
        header_alignment="left",
        clear=True,
        restore_index=False,
        ascending=False,
        _recursion=False,
        **kwargs,
    ):
        """Add items to the combobox's standard modelView without triggering any signals.

        Parameters:
            x (str/list/dict): A string, list of strings, or dict with 'string':data pairs to fill the comboBox with.
            data (optional): The data associated with the items.
            header (str, optional): An optional value for the first index of the comboBox's list.
            clear (bool, optional): Whether to clear any previous items before adding new. Defaults to True.
            restore_index (bool, optional): Whether to restore the previous index after clearing. Defaults to False.
            ascending (bool, optional): Whether to insert in ascending order. If True, new item(s) will be added to the top of the list. Defaults to False.
            _recursion (bool, optional): Internal use only. Differentiates between the initial call and recursive calls.
            kwargs: Arbitrary keyword arguments to set attributes for the added items.

        Returns:
            widget/list: The added widget or list of added widgets.

        Raises:
            TypeError: If the type of 'x' is unsupported.
        """
        self.blockSignals(True)

        last_added_data = None
        last_added_text = None

        def add_single(item, data):
            nonlocal last_added_data, last_added_text
            if ascending:
                self.insertItem(0, item, data)
            else:
                self.addItem(item, data)
            last_added_data = data
            last_added_text = item

        if not _recursion and clear:
            prev_index = self.currentIndex()
            self.clear()

        if isinstance(x, dict):
            [add_single(k, v) for k, v in x.items()]
        elif isinstance(x, (list, tuple, set)):
            [add_single(item, data) for item in x]
        elif isinstance(x, (zip, map)):
            [add_single(i, d) for i, d in x]
        elif isinstance(x, str):
            add_single(x, data)
        else:
            raise TypeError(
                f"Unsupported item type: '{type(x)}'. Expected str, list, tuple, set, map, zip, or dict."
            )

        self.set_attributes(**kwargs)

        if not _recursion:
            self.blockSignals(True)

            final_index = 0  # Default index is 0
            if restore_index and prev_index > 0:
                final_index = prev_index

            self.setCurrentIndex(final_index)

            if header:
                self.setHeader(header, header_alignment)

            self.blockSignals(False)
            self.currentIndexChanged.emit(final_index)

    @block_signals
    def removeItem(self, index=None):
        if index is None:
            index = self.currentIndex()
        item_text = self.itemText(index)
        super().removeItem(index)
        self.item_deleted.emit(item_text)

    @block_signals
    def currentData(self):
        return self.itemData(self.currentIndex())

    @block_signals
    def setCurrentData(self, value):
        self.setItemData(self.currentIndex(), value)

    @block_signals
    def currentText(self):
        return self.richText(self.currentIndex())

    @block_signals
    def setCurrentText(self, text):
        self.setRichText(text, self.currentIndex())

    @block_signals
    def setItemText(self, index, text):
        self.setRichText(text, index)

    def setCurrentItem(self, i, block_signals=False):
        if block_signals:
            self.blockSignals(True)

        try:
            index = (
                self.items.index(i)
                if isinstance(i, str)
                else i
                if isinstance(i, int)
                else None
            )
        except ValueError:
            raise ValueError(
                f"The item '{i}' was not found in ComboBox. Available items are {self.items}."
            )

        if index is None:
            raise RuntimeError(
                f"Failed to set current item in ComboBox: expected int or str, got {i, type(i)}"
            )

        self.setCurrentIndex(index)

        if block_signals:
            self.blockSignals(False)

    def setEditable(self, editable):
        if editable:
            current_text = self.currentText()
            super().setEditable(True)
            lineEdit = self.lineEdit()
            lineEdit.setText(current_text)
            lineEdit.deselect()
        else:
            lineEdit = self.lineEdit()
            new_text = lineEdit.text()
            super().setEditable(False)
            self.setCurrentText(new_text)
            self.on_editing_finished.emit(new_text)

    def showPopup(self):
        self.view().setMinimumWidth(self.sizeHint().width())
        self.before_popup_shown.emit()
        super().showPopup()

    def keyPressEvent(self, event):
        if self.isEditable() and event.key() == QtCore.Qt.Key_Return:
            self.setEditable(False)

        super().keyPressEvent(event)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    # return the existing QApplication object, or create a new one if none exists.
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    cmb = ComboBox()
    cmb.editable = True
    cmb.add(["Item A", "Item B"], header="Items:", header_alignment="center")

    cmb.show()
    sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace,
        then right-click them and select 'Promote to...'.

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote",
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""
